import java.io.*;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Hashtable;

class WholeOrbitData {

	WholeOrbitData(String wodFileName,String coeffFileName,String rudak) {
		this.rudak=rudak.toUpperCase();
		if(this.rudak.equals("A") || this.rudak.equals("B")) {
			try {
				loadCoefficients(coeffFileName);
				if(this.rudak.equals("A"))
					coefficients=RUDAKACoefficients;
				else
					coefficients=RUDAKBCoefficients;
				loadWholeOrbitData(wodFileName);
			} catch (FileNotFoundException e) {
				System.err.println(e.toString());
			} catch (EOFException e) {
				System.err.println(e.toString());
			} catch (IOException e) {
				System.err.println(e.toString());
			}
		} else {
			System.err.println(usage);
			System.exit(1);
		}
	}

	public static void main(String[] args) {
		if(args.length!=3) {
			System.err.println(usage);
			System.exit(1);
		}
		WholeOrbitData WOD=new WholeOrbitData(args[0],args[1],args[2]);
	}

	private SimpleDateFormat dateFormat=new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
	private DecimalFormat valueFormat=new DecimalFormat("#.##");
	private static final String usage="Usage: java WholeOrbitData <WOD File> <Coeff File> <A | B>";
	private String rudak;
	private final int hashSize=64;
	private Hashtable RUDAKACoefficients=new Hashtable(hashSize);
	private Hashtable RUDAKBCoefficients=new Hashtable(hashSize);
	private Hashtable coefficients;
	private RUDAKCoefficient wodChannels[];

	private void loadCoefficients(String coeffFileName)
		throws FileNotFoundException, IOException {

		final int COEFF_UNDEFINED=0;
		final int COEFF_RUDAKA=1;
		final int COEFF_RUDAKCANA=2;
		final int COEFF_RUDAKB=3;
		final int COEFF_RUDAKCANB=4;
		final int COEFF_NOTES=5;
		int state=COEFF_UNDEFINED;

		RUDAKCoefficient rc;
		CANCoefficient cc;
		BufferedReader in=new BufferedReader(new FileReader(coeffFileName));
		String line;
		while((line=in.readLine())!=null) {
			if(line.equals("RUDAKA")) {
				state=COEFF_RUDAKA;
			} else if(line.equals("RUDAKCANA")) {
				state=COEFF_RUDAKCANA;
			} else if(line.equals("RUDAKB")) {
				state=COEFF_RUDAKB;
			} else if(line.equals("RUDAKCANB")) {
				state=COEFF_RUDAKCANB;
			} else if(line.equals("NOTES")) {
				state=COEFF_NOTES;
			} else {
				switch(state) {
					case COEFF_UNDEFINED:
						break;
					case COEFF_RUDAKA:
						rc=new RUDAKCoefficient(line);
						RUDAKACoefficients.put(new Integer(rc.getChannel()),rc);
						break;
					case COEFF_RUDAKCANA:
						cc=new CANCoefficient(line);
						RUDAKACoefficients.put(new Integer(cc.getChannel()),cc);
						break;
					case COEFF_RUDAKB:
						rc=new RUDAKCoefficient(line);
						RUDAKBCoefficients.put(new Integer(rc.getChannel()),rc);
						break;
					case COEFF_RUDAKCANB:
						cc=new CANCoefficient(line);
						RUDAKBCoefficients.put(new Integer(cc.getChannel()),cc);
						break;
					case COEFF_NOTES:
						break;
				}
			}
		}

	}

	private void loadWholeOrbitData(String wodFileName)
		throws FileNotFoundException, IOException {

		DataInputStream in=new DataInputStream(
					new BufferedInputStream(
						new FileInputStream(wodFileName)));
		
		// get the start time
		// get the end time
		// get the interval
		// get the number of channels

		long start=(long)getIntelData(in,4);
		long finish=(long)getIntelData(in,4);
		int interval=getIntelData(in,2);
		int channels=getIntelData(in,1);

		long expected=(finish-start)/(long)interval;

		// output header to errout
		System.err.println("Start: "+dateFormat.format(new Date(start*1000)));
		System.err.println("Finish: "+dateFormat.format(new Date(finish*1000)));
		System.err.println("Expected Samples: "+expected);

		// also output to the csv file
		System.out.println("Start,"+dateFormat.format(new Date(start*1000)));
		System.out.println("Finish,"+dateFormat.format(new Date(finish*1000)));
		System.out.println("Expected Samples,"+expected);
		

		// for each channel
		//   get channel #
		wodChannels=new RUDAKCoefficient[channels];
		int channel=0;
		for(int i=0;i<channels;i++) {
			channel=getIntelData(in,1);
			wodChannels[i]=(RUDAKCoefficient)(coefficients.get(new Integer(channel)));
			if(wodChannels[i]==null) {
				wodChannels[i]=new UndefinedCoefficient(channel);
			}
		}

		// print the channel units
		for(int i=0;i<channels;i++) {
			System.out.print(",");
			System.out.print(wodChannels[i].getUnits());
		}
		System.out.println();

		// print the channel numbers 
		for(int i=0;i<channels;i++) {
			System.out.print(",");
			System.out.print(Integer.toHexString(wodChannels[i].getChannel()).toUpperCase());
		}
		System.out.println();

		// print the channel names (and date and time)
		System.out.print("Date/Time");
		for(int i=0;i<channels;i++) {
			System.out.print(",");
			System.out.print(wodChannels[i].getName());
		}
		System.out.println();


		// for each time interval
		//   for each channel
		//     get value

		int count=0;
		long current=start;
		try {
			while(current<=finish) {
				System.out.print(dateFormat.format(new Date(current*1000)));
				for(int i=0;i<channels;i++) {
					System.out.print(",");
					System.out.print(valueFormat.format(wodChannels[i].value(getIntelData(in,2))));
				}
				System.out.println();
				current+=interval;
				count++;
			}
		} catch (EOFException e) {
			System.err.println("Unexpected EOF on "+wodFileName);
			System.err.println("Expected: "+expected);
			System.err.println("Got: "+count);
			System.err.println("Last at: "+dateFormat.format(new Date(current*1000)));

			// also output to the csv file
			System.out.println();
			System.out.println("Unexpected EOF,"+wodFileName);
			System.out.println("Expected,"+expected);
			System.out.println("Got,"+count);
			System.out.println("Last at,"+dateFormat.format(new Date(current*1000)));

		}
	}

	private int getIntelData(DataInputStream in,int bytes)
		throws IOException {
		int value=0;
		int multiplier=1;
		for(int i=0;i<bytes;i++) {
			value+=in.readUnsignedByte()*multiplier;
			multiplier*=256;
		}
		return value;
	}

	class RUDAKCoefficient {

		int channel;
		String name;
		double a;
		double b;
		double c;
		String units;

		RUDAKCoefficient() {
		}

		RUDAKCoefficient(String line) {
			channel=Integer.parseInt(line.substring(0,2).trim(),16);
			name=line.substring(6,20).trim();
			c=Double.parseDouble(line.substring(21,28).trim());
			b=Double.parseDouble(line.substring(34,41).trim());
			a=Double.parseDouble(line.substring(46,57).trim());
			units=line.substring(58,line.length()).trim();
		}

		int getChannel() {
			return channel;
		}

		String getName() {
			return name;
		}

		String getUnits() {
			return units;
		}

		double value(int x) {
			return (a*(double)x*(double)x)+(b*(double)x)+c;
		}

	}

	class CANCoefficient extends RUDAKCoefficient {

		double d;
		double e;
		double f;
		
		CANCoefficient(String line) {
			channel=Integer.parseInt(line.substring(0,2).trim(),16);
			name=line.substring(6,20).trim();
			a=Double.parseDouble(line.substring(21,37).trim());
			b=Double.parseDouble(line.substring(38,52).trim());
			c=Double.parseDouble(line.substring(53,67).trim());
			d=Double.parseDouble(line.substring(68,79).trim());
			e=Double.parseDouble(line.substring(80,91).trim());
			f=Double.parseDouble(line.substring(92,101).trim());
			units=line.substring(102,line.length()).trim();
		}

		double value(int x) {
			return (a*Math.pow((double)x,5.0)) +
				(b*Math.pow((double)x,4.0)) +
				(c*Math.pow((double)x,3.0)) +
				(d*Math.pow((double)x,2.0)) +
				(e*(double)x) +
				f;
		}
	}


	class UndefinedCoefficient extends RUDAKCoefficient {

		UndefinedCoefficient(int channel) {
			this.channel=channel;
			name="Undefined";
			units="Undefined";
		}

		double value(int x) {
			return (double)x;
		}
	}

}
